/**
 * Copyright 2022 Google LLC
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *      https://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.cast.sample.casthelloaudioandroid;

import androidx.annotation.NonNull;
import androidx.appcompat.app.AppCompatActivity;

import android.annotation.SuppressLint;
import android.content.ComponentName;
import android.content.Intent;
import android.graphics.Bitmap;
import android.media.AudioManager;
import android.os.Bundle;
import android.support.v4.media.MediaBrowserCompat;
import android.support.v4.media.MediaMetadataCompat;
import android.support.v4.media.session.MediaControllerCompat;
import android.support.v4.media.session.MediaSessionCompat;
import android.support.v4.media.session.PlaybackStateCompat;
import android.util.Log;
import android.view.Menu;
import android.view.View;
import android.widget.Button;
import android.widget.ImageButton;
import android.widget.ImageView;
import android.widget.SeekBar;
import android.widget.TextView;
import android.widget.Toast;

import com.google.android.gms.cast.CastStatusCodes;
import com.google.android.gms.cast.framework.CastButtonFactory;
import com.google.android.gms.cast.framework.CastContext;
import com.google.android.gms.cast.framework.CastSession;
import com.google.android.gms.cast.framework.SessionManager;
import com.google.android.gms.cast.framework.SessionManagerListener;

import java.util.List;

public class MainActivity extends AppCompatActivity {

    private ImageView albumArt;
    private ImageButton playButton;
    private ImageButton pauseButton;
    private ImageButton fwdButton;
    private ImageButton rwdButton;
    private ImageButton stopButton;
    private Button loadAudioButton;
    private SeekBar audioSeekbar;
    private TextView bufferingText;

    private CastSession castSession;
    private CastContext castContext;
    private SessionManager sessionManager;
    private SessionManagerListener<CastSession> sessionManagerListener =
            new MySessionManagerListener();

    private final String LOG_TAG = "**MainActivity**";

    /** Currently not/partially in use, WIP */
    public static final String AUTO_PLAY = "com.google.cast.sample.casthelloaudioandroid.AUTOPLAY";
    public static final String KILL_COMMAND = "com.google.cast.sample.casthelloaudioandroid.STOP_APP";
    public static final String CURR_POSITION = "com.google.cast.sample.casthelloaudioandroid.CURRENT_POSITION";
    private PlaybackLocation playbackLocation;

    private MediaBrowserCompat mediaBrowser;

    public enum PlaybackLocation {
        LOCAL,
        REMOTE
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        albumArt = findViewById(R.id.albumArt);

        playButton = findViewById(R.id.playbutton);

        pauseButton = findViewById(R.id.pauseButton);

        fwdButton = findViewById(R.id.fwdButton);
        fwdButton.setVisibility(View.GONE);
        //TODO - Implement Skip forward 30 seconds

        rwdButton = findViewById(R.id.rwdButton);
        rwdButton.setVisibility(View.GONE);
        //TODO - Implement Skip back 30 seconds

        stopButton = findViewById(R.id.stopButton);

        loadAudioButton = findViewById(R.id.loadAudioButton);
        loadAudioButton.setVisibility(View.INVISIBLE);

        audioSeekbar = findViewById(R.id.audioSeekbar);
        audioSeekbar.setVisibility(View.INVISIBLE);
        //TODO - Implement setOnSeekbarChangedListener

        bufferingText = findViewById(R.id.bufferingText);

        castContext = CastContext.getSharedInstance(this);

        sessionManager = castContext.getSessionManager();
        sessionManager.addSessionManagerListener(sessionManagerListener, CastSession.class);

        startMediaBrowserService();
    }

    private void startMediaBrowserService() {
        mediaBrowser = new MediaBrowserCompat(this,
                new ComponentName(this, AudioPlayerService.class),
                audioServiceConnectionCallbacks,
                null);
        mediaBrowser.subscribe("root", new MediaBrowserCompat.SubscriptionCallback() {
            @Override
            public void onChildrenLoaded(@NonNull String parentId, @NonNull List<MediaBrowserCompat.MediaItem> children) {
                super.onChildrenLoaded(parentId, children);
                showToast("Children Loaded", Toast.LENGTH_SHORT);
            }
        });
    }

    @Override
    protected void onStart() {
        super.onStart();
        if (!mediaBrowser.isConnected()) {
            mediaBrowser.connect();
        }
    }

    @Override
    protected void onResume() {
        super.onResume();
        if (castSession == null) {
            playbackLocation = PlaybackLocation.LOCAL;
        } else {
            playbackLocation = PlaybackLocation.REMOTE;
            if (castSession.getRemoteMediaClient().hasMediaSession()) {
                setPlayerImage();
            }
        }
        setVolumeControlStream(AudioManager.STREAM_MUSIC);
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
        if (castContext != null) {
            sessionManager.removeSessionManagerListener(sessionManagerListener, CastSession.class);
            castContext = null;
        }
        castSession = null;

        if (MediaControllerCompat.getMediaController(MainActivity.this) != null) {
            MediaControllerCompat.getMediaController(MainActivity.this).unregisterCallback(controllerCallback);
        }
        mediaBrowser.disconnect();
        mediaBrowser.unsubscribe("root");
        Log.d(LOG_TAG, "onDestroy Called");
    }

    private final MediaBrowserCompat.ConnectionCallback audioServiceConnectionCallbacks =
            new MediaBrowserCompat.ConnectionCallback() {
                @Override
                public void onConnected() {
                    MediaSessionCompat.Token token = mediaBrowser.getSessionToken();
                    MediaControllerCompat mediaController =  new MediaControllerCompat(MainActivity.this, token);
                    MediaControllerCompat.setMediaController(MainActivity.this, mediaController);
                    buildTransportControls();
                }

                @Override
                public void onConnectionSuspended() {
                    super.onConnectionSuspended();
                    Log.d(LOG_TAG, "**audioServiceConnectionCallbacks: onConnectionSuspended");
                }

                @Override
                public void onConnectionFailed() {
                    super.onConnectionFailed();
                    Log.d(LOG_TAG, "**audioServiceConnectionCallbacks: onConnectionFailed");
                }
            };

    private void buildTransportControls() {
        MediaControllerCompat.TransportControls transportControls =
                MediaControllerCompat.getMediaController(MainActivity.this).getTransportControls();
        transportControls.prepare();

        playButton.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View v) {
                transportControls.play();
            }
        });

        pauseButton.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View v) {
                transportControls.pause();
            }
        });

        stopButton.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View v) {
                transportControls.stop();
            }
        });

        MediaControllerCompat mediaController = MediaControllerCompat.getMediaController(MainActivity.this);
        mediaController.registerCallback(controllerCallback);
    }

    private MediaControllerCompat.Callback controllerCallback = new MediaControllerCompat.Callback() {
        @Override
        public void onSessionReady() {
            super.onSessionReady();
            setPlayerImage();
        }

        @Override
        public void onSessionDestroyed() {
            mediaBrowser.disconnect();
        }

        @SuppressLint("RestrictedApi")
        @Override
        public void onPlaybackStateChanged(PlaybackStateCompat state) {
            super.onPlaybackStateChanged(state);
            showToast("PlaybackStateChanged: "+state.toString()+" "+state.getCurrentPosition(null), Toast.LENGTH_SHORT);

        }

        @Override
        public void onMetadataChanged(MediaMetadataCompat metadata) {
            super.onMetadataChanged(metadata);
            showToast("MetadataChanged: "+metadata.toString(), Toast.LENGTH_SHORT);
        }
    };

    @Override
    protected void onNewIntent(Intent intent) {
        super.onNewIntent(intent);
        showToast("onNewIntent called", Toast.LENGTH_SHORT);
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        super.onCreateOptionsMenu(menu);
        getMenuInflater().inflate(R.menu.menu_main, menu);
        CastButtonFactory.setUpMediaRouteButton(getApplicationContext(),
                menu,
                R.id.media_route_menu_item);
        return true;
    }

    private void checkImageURI() throws NullPointerException{
        if (AudioPlayerService.KEY_IMAGE_URI == null) {
            throw new NullPointerException("KEY_IMAGE_URI is null and not set in AudioPlayerService.class. \n" +
                    "Check and complete TODOs mentioned in README.md");
        }
    }

    public void setPlayerImage() {
        checkImageURI();
        new Thread(new Runnable() {
            @Override
            public void run() {
                final Bitmap img = new BitmapImageProvider(AudioPlayerService.KEY_IMAGE_URI).fetchBitmap();
                albumArt.post(new Runnable() {
                    @Override
                    public void run() {
                        albumArt.setImageBitmap(img);
                    }
                });
            }
        }).start();
    }

    private void showToast(String msg, int length) {
        Toast.makeText(getApplicationContext(), msg, length).show();
    }

    private class MySessionManagerListener implements SessionManagerListener<CastSession> {

        @Override
        public void onSessionEnded(@NonNull @org.jetbrains.annotations.NotNull CastSession session, int i) {
            Log.d(LOG_TAG, "Session Ended: "+session.getSessionId()+" Reason: "
                    +CastStatusCodes.getStatusCodeString(i));
            if (castSession == session) {
                castSession = null;
            }
            playbackLocation = PlaybackLocation.LOCAL;
            invalidateOptionsMenu();
        }

        @Override
        public void onSessionEnding(@NonNull @org.jetbrains.annotations.NotNull CastSession session) {
            Log.d(LOG_TAG, "Session Ending: "+session.getSessionId());
        }

        @Override
        public void onSessionResumeFailed(@NonNull @org.jetbrains.annotations.NotNull CastSession session, int i) {
            Log.e(LOG_TAG, "Session Resume Failed: "+CastStatusCodes.getStatusCodeString(i));
        }

        @Override
        public void onSessionResumed(@NonNull @org.jetbrains.annotations.NotNull CastSession session, boolean b) {
            Log.d(LOG_TAG, "Session Resumed: "+b);
            castSession = session;
            castSession.getCastDevice();
            playbackLocation = PlaybackLocation.REMOTE;
            invalidateOptionsMenu();
        }

        @Override
        public void onSessionResuming(@NonNull @org.jetbrains.annotations.NotNull CastSession session, @NonNull @org.jetbrains.annotations.NotNull String s) {
            Log.d(LOG_TAG, "Session Resuming: "+s);
        }

        @Override
        public void onSessionStartFailed(@NonNull @org.jetbrains.annotations.NotNull CastSession session, int i) {
            Log.d(LOG_TAG, "Session Start Failed: "+ CastStatusCodes.getStatusCodeString(i));
        }

        @Override
        public void onSessionStarted(@NonNull @org.jetbrains.annotations.NotNull CastSession session, @NonNull @org.jetbrains.annotations.NotNull String s) {
            Log.d(LOG_TAG, "Session Started: "+session.getSessionId()+" s: "+s);
            castSession = session;
            playbackLocation = PlaybackLocation.REMOTE;
            invalidateOptionsMenu();
        }

        @Override
        public void onSessionStarting(@NonNull @org.jetbrains.annotations.NotNull CastSession session) {
            Log.d(LOG_TAG, "Session Starting: "+session.getSessionId());
        }

        @Override
        public void onSessionSuspended(@NonNull @org.jetbrains.annotations.NotNull CastSession session, int i) {
            Log.e(LOG_TAG, "Session Suspended: "+CastStatusCodes.getStatusCodeString(i));
        }
    }
}